import pandas as pd
import io
import random

from colibrie.utils import (
    get_text_in_cell,
    extract_text_from_spans,
)

from PIL import Image, ImageDraw

from enum import Enum


class TableList(object):
    """Defines a list of core.Table objects. Each table can
    be accessed using its index.

    Attributes
    ----------
    n : int
        Number of tables in the list.

    """

    def __init__(self, tables):
        self._tables = tables

    def __repr__(self):
        return f"<{self.__class__.__name__} n={self.n}>"

    def __len__(self):
        return len(self._tables)

    def __getitem__(self, idx):
        return self._tables[idx]

    @property
    def n(self):
        return len(self)


class Table(object):
    """Defines a table with coordinates relative to a left-bottom
    origin. (PDF coordinate space)

    Parameters
    ----------
    cols : list
        List of tuples representing column x-coordinates in increasing
        order.
    rows : list
        List of tuples representing row y-coordinates in decreasing
        order.

    Attributes
    ----------
    df : :class:`pandas.DataFrame`
    shape : tupl
        Shape of the table.
    accuracy : float
        Accuracy with which text was assigned to the cell.
    whitespace : float
        Percentage of whitespace in the table.
    order : int
        Table number on PDF page.
    page : int
        PDF page number.

    """

    def __init__(self, intersections, rect, horizontal_lines, vertical_lines):
        self.rect = rect
        self.intersections = intersections

        self.intersections_df = pd.DataFrame(
            [[point.x, point.y] for point in self.intersections], columns=["x", "y"]
        )

        self.horizontal_lines = horizontal_lines
        self.vertical_lines = vertical_lines

        self.row = {
            row: index
            for index, row in enumerate(
                sorted(self.intersections_df.y.unique().tolist())
            )
        }
        self.col = {
            col: index
            for index, col in enumerate(
                sorted(self.intersections_df.x.unique().tolist())
            )
        }

        self.cells = [[] for r in range(len(self.row) - 1)]

        self.debug = Debug(self)

    def set_rotation(self):
        if self.rotation == Rotation.rotated_90.value:
            # Transpose table cell
            self.cells = list(zip(*reversed(self.cells)))

            # Invert row and col size which is a consequence of the transpose
            for row in self.cells:
                for cell in row:
                    if cell:
                        cell.row_size, cell.col_size = cell.col_size, cell.row_size

    def match_row_size_sum(self):
        # to permit the transposition of cells in order to rotate the table if needed
        # we append None value to match the size of sum(row_size) for each row
        try:
            for col_index in range(
                len(max(self.cells, key=len))
            ):  # this row get the maximum lenght of inner list in cells 2d array
                for row_index, row in enumerate(self.cells):
                    cell = row[col_index]
                    if cell:
                        for i in range(1, cell.row_size):
                            self.cells[row_index + i].insert(col_index, None)
        except IndexError:
            pass

    def to_html(self):
        import random

        # transposed_cells = list(zip(*self.cells)) # 180 degree turn
        # transposed_cells = list(zip(*[reversed(row) for row in self.cells])) # 270 degree turn
        html = "<table>"
        for row in self.cells:
            html += "<tr>"
            for cell in row:
                if cell:
                    rowspan = f"rowspan={cell.row_size}"
                    colspan = f"colspan={cell.col_size}"
                    html += f'<td style="text-align:center;" {rowspan} {colspan}>{cell.text}</td>'
            html += "</tr>"
        html += "</table>"

        return html

    def to_df(self):
        return pd.read_html(self.to_html())[0]


class Cell(object):
    """Defines a cell in a table with coordinates relative to a
    left-bottom origin. (PDF coordinate space)

    Parameters
    ----------
    rect : fitz.Rect object
    height : height in row count
    widht : widht in column count
    text : string
        Text assigned to cell.

    """

    def __init__(self, rect, row_size, col_size):
        self.rect = rect
        self.spans = None
        self.row_size = row_size
        self.col_size = col_size
        self.text = None

    def get_text(self, text):
        self.spans = get_text_in_cell(rect=self.rect, text=text)
        self.text = extract_text_from_spans(self.spans)

    def __repr__(self):
        return repr(self.text)


class Rotation(Enum):
    unrotated = (1.0, 0.0)
    rotated_90 = (0.0, -1.0)


class Debug:
    def __init__(self, table: Table, img_bytes: bytes = None):
        self.table = table
        self.image = img_bytes

    def show_cells(self, as_bytes: bool = True, image: bytes = None):
        if not self.image:
            print("Please ensure that the debug mode is activated")
            return

        with Image.open(io.BytesIO(image if image else self.image)) as img:
            draw = ImageDraw.Draw(img)

            for row in self.table.cells:
                for cell in row:
                    if cell:
                        color = (
                            int(random.uniform(30, 255)),
                            int(random.uniform(30, 255)),
                            int(random.uniform(30, 255)),
                        )
                        draw.rectangle(
                            [cell.rect.x0, cell.rect.y0, cell.rect.x1, cell.rect.y1],
                            outline=color,
                            width=2,
                        )

            if as_bytes:
                imgByteArr = io.BytesIO()
                img.save(imgByteArr, format=img.format)
                imgByteArr = imgByteArr.getvalue()

                return imgByteArr
            else:
                img.show()

    def show_border(self, as_bytes: bool = True, image: bytes = None):
        if not self.image:
            print("Please ensure that the debug mode is activated")
            return

        with Image.open(io.BytesIO(image if image else self.image)) as img:
            draw = ImageDraw.Draw(img)

            draw.rectangle(
                [
                    self.table.rect.x0,
                    self.table.rect.y0,
                    self.table.rect.x1,
                    self.table.rect.y1,
                ],
                outline=(0, 128, 0),
                width=2,
            )

            if as_bytes:
                imgByteArr = io.BytesIO()
                img.save(imgByteArr, format=img.format)
                imgByteArr = imgByteArr.getvalue()

                return imgByteArr
            else:
                img.show()

    def show_intersections(self, as_bytes: bool = True, image: bytes = None):
        if not self.image:
            print("Please ensure that the debug mode is activated")
            return

        with Image.open(io.BytesIO(image if image else self.image)) as img:
            draw = ImageDraw.Draw(img)

            for intersection in self.table.intersections:
                tl = (intersection.x - 3, intersection.y - 3)
                br = (intersection.x + 3, intersection.y + 3)
                coordinate = [tl, br]
                draw.ellipse(coordinate, outline=(255, 0, 0), width=2)

            if as_bytes:
                imgByteArr = io.BytesIO()
                img.save(imgByteArr, format=img.format)
                imgByteArr = imgByteArr.getvalue()

                return imgByteArr
            else:
                img.show()

    def show_lines(self, as_bytes: bool = True, image: bytes = None):
        if not self.image:
            print("Please ensure that the debug mode is activated")
            return

        with Image.open(io.BytesIO(image if image else self.image)) as img:
            draw = ImageDraw.Draw(img)

            lines = self.table.horizontal_lines + self.table.vertical_lines

            for line in lines:
                color = (
                    int(random.uniform(30, 255)),
                    int(random.uniform(30, 255)),
                    int(random.uniform(30, 255)),
                )
                draw.line(
                    [line[0].x, line[0].y, line[1].x, line[1].y], fill=color, width=2
                )

            if as_bytes:
                imgByteArr = io.BytesIO()
                img.save(imgByteArr, format=img.format)
                imgByteArr = imgByteArr.getvalue()

                return imgByteArr
            else:
                img.show()



from fitz import Rect, TEXTFLAGS_DICT, TEXT_PRESERVE_IMAGES

from colibrie.geometry import Point
from colibrie.core import Table, Cell, Rotation
from ailist import AIList


def get_tables_candidates(vertical_lines, horizontal_lines):
    interval_tree = AIList()
    [interval_tree.add(line[0].y, line[1].y) for line in vertical_lines]

    vertical_merged_interval = interval_tree.merge(gap=1)

    tables_candidate = []
    interval_group = {}

    for vertical_interval in vertical_merged_interval:
        interval_group[vertical_interval.start, vertical_interval.end] = []

        for line in vertical_lines:
            if (
                line[0].y >= vertical_interval.start
                and line[1].y <= vertical_interval.end
            ):
                interval_group[vertical_interval.start, vertical_interval.end].append(
                    line
                )

        for line in horizontal_lines:
            if (
                line[0].y >= vertical_interval.start
                and line[1].y <= vertical_interval.end
            ):
                interval_group[vertical_interval.start, vertical_interval.end].append(
                    line
                )

        # get horizontal interval from each group
        interval_tree = AIList()
        [
            interval_tree.add(line[0].x, line[1].x)
            for line in interval_group[vertical_interval.start, vertical_interval.end]
        ]

        horizontal_merged_interval = interval_tree.merge(gap=1)

        for horizontal_interval in horizontal_merged_interval:
            table_candidate = []

            for line in interval_group[vertical_interval.start, vertical_interval.end]:
                if (
                    line[0].x >= horizontal_interval.start
                    and line[1].x <= horizontal_interval.end
                ):
                    table_candidate.append(line)

            # We make sure that every table candidate have at least 4 lines
            # Which is the bare minimum to create a single cell
            if len(table_candidate) > 3:
                tables_candidate.append(table_candidate)
            else:
                continue

    return tables_candidate


def create_table(intersections, horizontal_lines, vertical_lines):
    intersections = [
        intersection
        for intersection in sorted(intersections, key=lambda point: (point.y, point.x))
    ]

    table = Table(
        intersections=intersections,
        rect=Rect(intersections[0], intersections[-1]),
        horizontal_lines=horizontal_lines,
        vertical_lines=vertical_lines,
    )

    return table


def process_table(page, table, preserve_span, intersections_info):
    """
    tl => top left
    tr => top right
    bl => bottom left
    br => bottom right
    """

    rotated_span = {}
    rotated_span[Rotation.unrotated.value] = 0
    rotated_span[Rotation.rotated_90.value] = 0

    if preserve_span:
        text = page.get_text(
            "dict", flags=TEXTFLAGS_DICT & ~TEXT_PRESERVE_IMAGES, clip=table.rect
        )
    else:
        textpage = page.get_textpage(
            flags=TEXTFLAGS_DICT & ~TEXT_PRESERVE_IMAGES, clip=table.rect
        )

    # Reconstruct intersection point group by lines height as
    # list of list containing intersections point
    rows = []
    for index, row in table.intersections_df.groupby("y"):
        rows.append(row.to_dict(orient="records"))

    for row_index, row in enumerate(rows):
        # Stopping one row before the last row because
        # the last row is always the bottom of the table
        # and can't be used for a cell matching
        if row_index == len(rows) - 1:
            break

        tl = None
        for intersection_index, intersection in enumerate(row):
            if intersection_index == len(row) - 1:
                break

            if not tl:
                tl = intersection

            tr = row[intersection_index + 1]
            tl_intersection_line = intersections_info[tl["x"], tl["y"]]
            tr_intersection_line = intersections_info[tr["x"], tr["y"]]

            # if tl and tr point are the same horizontal_line
            if tl_intersection_line[1] == tr_intersection_line[1]:
                bl, br = find_intersection_correspondance(
                    tl, tr, rows, row_index, intersections_info
                )
            else:
                tl = None
                continue

            if not bl and not br:
                continue

            cell = Cell(
                rect=Rect(Point(tl["x"], tl["y"]), Point(br["x"], br["y"])),
                row_size=table.row[br["y"]] - table.row[tl["y"]],
                col_size=table.col[br["x"]] - table.col[tl["x"]],
            )

            if preserve_span:
                cell.get_text(text)
                for span in cell.spans:
                    try:
                        rotated_span[span["dir"]] += 1
                    except KeyError:
                        print(f"missing key for rotation {span['dir']}")
            else:
                cell.text = page.get_textbox(cell.rect, textpage=textpage)

            table.cells[row_index].append(cell)

            # to permit the transposition of cells in order to rotate the table if needed
            # we append None value to match the size of sum(col_size) for each row
            for i in range(cell.row_size):
                for j in range(1, cell.col_size):
                    table.cells[row_index + i].append(None)

            tl = None

    ############## HANDLE BAD TABLE ######################
    total_cell = 0
    empty_cell = 0
    for row in table.cells:
        for cell in row:
            if cell:
                total_cell += 1
                if not cell.text:
                    empty_cell += 1

    empty_cell_ratio = total_cell / max(
        empty_cell, 1
    )  # prevent division by 0 if empty table

    # if there is only empty cell in the table, it is not a valid table
    if empty_cell_ratio == 1:
        return None
    #######################################################
    else:
        table.match_row_size_sum()
        table.rotation = max(rotated_span, key=rotated_span.get)
        table.set_rotation()

    return table


def find_intersection_correspondance(
    tl, tr, intersection_rows, row_index, intersections_info
):
    """
    tl => top left
    tr => top right
    bl => bottom left
    br => bottom right
    """
    bl = None
    br = None

    # Starting to search for corresponding bl / br at row_index + 1 because we always
    # match a row with the row below it
    for row in intersection_rows[row_index + 1 :]:
        for intersection in row:

            if tl["x"] <= intersection["x"] < tr["x"] and not bl:
                bl = intersection

            if intersection["x"] == tr["x"] and not br and bl:
                br = intersection

            if br and bl:
                tl_intersection_line = intersections_info[tl["x"], tl["y"]]
                tr_intersection_line = intersections_info[tr["x"], tr["y"]]

                bl_intersection_line = intersections_info[bl["x"], bl["y"]]
                br_intersection_line = intersections_info[br["x"], br["y"]]

                if (
                    bl_intersection_line[1] == br_intersection_line[1]
                    and tl_intersection_line[  # if bl and br point are on the same horizontal_line
                        0
                    ]
                    == bl_intersection_line[0]
                    and tr_intersection_line[  # if tl and bl point are on the same vertical_line
                        0
                    ]
                    == br_intersection_line[0]
                    # if tr and br point are on the same vertical_line
                ):
                    return bl, br
                else:
                    bl = None
                    br = None

        # If we cannot find any valid candidate for closing the cell on this row, we need
        # to reset the start / end point value for the next row to None to prevent
        # Mixed row matching
        bl = None
        br = None

    return None, None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['colibrie']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.32,<0.30.0',
 'Pillow>=9.2.0,<10.0.0',
 'PyMuPDF>=1.20.2,<2.0.0',
 'ailist>=1.0.4,<2.0.0',
 'numpy>=1.23.3,<2.0.0',
 'pandas>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'colibrie',
    'version': '1.1.2.1',
    'description': 'Colibrie is a blazing fast tool to extract tables from PDFs',
    'long_description': '# Colibrie\n [![image](https://img.shields.io/pypi/v/colibrie.svg)](https://pypi.org/project/colibrie/) [![image](https://img.shields.io/pypi/l/colibrie.svg)](https://pypi.org/project/colibrie/)\n\nColibrie is a blazing fast tool to extract tables from PDFs \n\n## Why Colibrie?\n\n- :rocket: **Efficient**: Colibrie is faster by multiple order of magnitude than any actual existing solution\n- :sparkles: **Fidel visual**: Colibrie can provide 1:1 HTML representation of any tables it\'ll find\n- :books: **Reliable**: Colibri will find every valid tables without exception if the PDF is compatible with the core principle of Colibrie\n- :memo: **Output**: Each table can be export into to multiple formats, which include : \n  - Pandas Dataframe.\n  - HTML.\n\n### Benchmark :\nSome number to compare [Camelot](https://github.com/camelot-dev/camelot) (a popular library to extract tables from PDF) and Colibrie\n<table>\n  <thead>\n    <tr>\n        <th colspan="2"></th>\n        <th colspan="4">Tables extracted</th>\n        <th colspan="2"></th>\n    </tr>\n    <tr>\n        <th colspan="2">Times in second</th>\n        <th colspan="2">camelot</th>\n        <th colspan="2">colibrie</th>\n        <th colspan="2"></th>\n    </tr>\n    <tr style="text-align: right;">\n      <th>camelot</th>\n      <th>colibrie</th>\n      <th>valid</th>\n      <th>false positive</th>\n      <th>valid</th>\n      <th>false positive</th>\n      <th>pages count</th>\n      <th>pdf file</th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr>\n      <td>0.53</td>\n      <td>0.00545</td>\n      <td>1</td>\n      <td>0</td>\n      <td>1</td>\n      <td>0</td>\n      <td>1</td>\n      <td><a href="https://github.com/abitoun-42/colibrie/files/9620468/boc_20220014_0001_p000_extract_2.pdf">small pdf</a></td>\n    </tr>\n    <tr>\n      <td>5.95</td>\n      <td>0.02100</td>\n      <td>4</td>\n      <td>0</td>\n      <td>4</td>\n      <td>0</td>\n      <td>11</td>\n      <td><a href="https://github.com/abitoun-42/colibrie/files/9620506/boc_20210034_0000_0003.pdf">medium pdf</a></td>\n    </tr>\n    <tr>\n      <td>105.00</td>\n      <td>0.21900</td>\n      <td>62</td>\n      <td>1</td>\n      <td>62</td>\n      <td>0</td>\n      <td>167</td>\n      <td><a href="https://github.com/abitoun-42/colibrie/files/9620511/boc_20220014_0001_p000.pdf">big pdf</a></td>\n    </tr>\n    <tr>\n      <td>182.00</td>\n      <td>0.69000</td>\n      <td>175</td>\n      <td>1</td>\n      <td>177</td>\n      <td>0</td>\n      <td>269</td>\n      <td><a href="https://github.com/abitoun-42/colibrie/files/9620515/boc_20220025_0001_p000.pdf">giant pdf</a></td>\n    </tr>\n  </tbody>\n</table>\n\n## Installation\n\n### using source\n```\npip install poetry\n\ngit clone https://github.com/abitoun-42/colibrie.git\n\ncd colibrie\n\npoetry install\n```\n### using pip\n```\npip install colibrie\n```\n\n## Usage\n```\nfrom colibrie.extract_tables import extract_table\n\ntables = extract_table(\'pdf_path\')\n```\n',
    'author': 'Alexandre Bitoun',
    'author_email': 'alexandre.bitoun@outlook.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/abitoun-42/colibrie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
